import typing

import sqlalchemy as sa

class AbstractEntity:

    metadata: typing.ClassVar[sa.MetaData]

    def __init__(self, **kwargs: typing.Any) -> None: ...
    @classmethod
    def __tablename__(cls) -> str: ...
    @classmethod
    def classname(cls) -> str: ...

class Entity(AbstractEntity):
    id_: sa.Column[sa.Integer]

    @property
    def pk(self) -> sa.Column[sa.Integer]: ...

def make_table(
    name: str, *args: sa.schema.SchemaItem, **kwargs: typing.Any
) -> sa.Table: ...



import contextlib
from inspect import isclass
from typing import cast

from sqlalchemy.sql import ColumnElement
from typing_extensions import TypeGuard

from gyver.database.entity import AbstractEntity

from .exc import FieldNotFound
from .interface import FieldType
from .interface import Mapper


def retrieve_attr(entity: Mapper, field: str) -> FieldType:
    is_entity = _is_entity(entity)
    if field == "id" and is_entity:
        field = "id_"
    if "." in field:
        if is_entity:
            return _retrieve_related_field(entity, field)
        raise FieldNotFound("query", field)
    try:
        with contextlib.suppress(AttributeError):
            return getattr(entity, field)
        return getattr(entity.c, field)  # type: ignore
    except AttributeError:
        name = entity.__name__ if is_entity else "query"
        raise FieldNotFound(name, field) from None


def _is_entity(entity: Mapper) -> TypeGuard[type[AbstractEntity]]:
    return isclass(entity) and issubclass(entity, AbstractEntity)


def _retrieve_related_field(entity: type[AbstractEntity], field: str) -> FieldType:
    *fields, target_field = field.split(".")
    current_mapper = entity
    for f in fields:
        if f == "id":
            f = "id_"
        try:
            attr = cast(ColumnElement, getattr(current_mapper, f))
        except AttributeError:
            raise FieldNotFound(entity.__name__, f) from None
        else:
            current_mapper = attr.entity.class_
    if target_field == "id":
        target_field = "id_"
    try:
        return getattr(current_mapper, target_field)
    except AttributeError:
        raise FieldNotFound(entity.__name__, field) from None



import typing

from gyver.context.typedef import T


class Adapter(typing.Protocol[T]):
    def is_closed(self, client: T) -> bool:
        """Returns if client state is closed or released"""
        ...

    def release(self, client: T) -> None:
        """Closes or releases client"""
        ...

    def new(self) -> T:
        """Creates new client"""
        ...


class AsyncAdapter(typing.Protocol[T]):
    """Represents an Adapter to a Specific Service"""

    async def is_closed(self, client: T) -> bool:
        """Returns if client state is closed or released"""
        ...

    async def release(self, client: T) -> None:
        """Closes or releases client"""
        ...

    async def new(self) -> T:
        """Creates new client"""
        ...


class AtomicAdapter(Adapter[T], typing.Protocol[T]):
    """Represents an Adapter to a Specific Service that
    can do atomic operations"""

    def begin(self, client: T) -> None:
        ...

    def commit(self, client: T) -> None:
        ...

    def rollback(self, client: T) -> None:
        ...

    def in_atomic(self, client: T) -> bool:
        ...


class AtomicAsyncAdapter(AsyncAdapter[T], typing.Protocol[T]):
    """Represents an Adapter to a Specific Service that
    can do atomic operations"""

    async def begin(self, client: T) -> None:
        ...

    async def commit(self, client: T) -> None:
        ...

    async def rollback(self, client: T) -> None:
        ...

    async def in_atomic(self, client: T) -> bool:
        ...



import typing

import sqlalchemy.ext.asyncio as sa_asyncio

from gyver import context
from gyver.utils import lazyfield

AsyncSaContext = context.AsyncContext[sa_asyncio.AsyncConnection]

AsyncSessionContext = context.AsyncContext[sa_asyncio.AsyncSession]


class AsyncConnectionAdapter(context.AtomicAsyncAdapter[sa_asyncio.AsyncConnection]):
    @typing.overload
    def __init__(
        self,
        *,
        uri: str,
        engine: None = None,
    ) -> None:
        ...

    @typing.overload
    def __init__(
        self,
        *,
        uri: None = None,
        engine: sa_asyncio.AsyncEngine,
    ) -> None:
        ...

    def __init__(
        self,
        *,
        uri: typing.Optional[str] = None,
        engine: typing.Optional[sa_asyncio.AsyncEngine] = None,
    ) -> None:
        if not any((uri, engine)):
            raise TypeError("Missing parameters (uri/engine)")
        self._uri = uri
        if engine is not None:
            self._engine = engine

    @lazyfield
    def _engine(self):
        return sa_asyncio.create_async_engine(self._uri)

    def _create_connection(self):
        return self._engine.connect()

    async def is_closed(self, client: sa_asyncio.AsyncConnection) -> bool:
        return client.closed

    async def release(self, client: sa_asyncio.AsyncConnection) -> None:
        return await client.close()

    async def new(self) -> sa_asyncio.AsyncConnection:
        return await self._create_connection()

    async def begin(self, client: sa_asyncio.AsyncConnection) -> None:
        if client.in_transaction():
            await client.begin_nested()
        else:
            await client.begin()

    async def commit(self, client: sa_asyncio.AsyncConnection) -> None:
        if trx := (client.get_nested_transaction() or client.get_transaction()):
            await trx.commit()

    async def rollback(self, client: sa_asyncio.AsyncConnection) -> None:
        if trx := (client.get_nested_transaction() or client.get_transaction()):
            await trx.rollback()

    async def in_atomic(self, client: sa_asyncio.AsyncConnection) -> bool:
        return client.in_transaction()

    def context(
        self,
    ) -> "AsyncSaContext":
        return AsyncSaContext(self)


class AsyncSessionAdapter(context.AtomicAsyncAdapter[sa_asyncio.AsyncSession]):
    def __init__(
        self, adapter: context.AtomicAsyncAdapter[sa_asyncio.AsyncConnection]
    ) -> None:
        self._internal_adapter = adapter

    async def new(self):
        return sa_asyncio.AsyncSession(await self._internal_adapter.new())

    async def release(self, client: sa_asyncio.AsyncSession) -> None:
        await client.close()
        await self._internal_adapter.release(client.bind)

    async def is_closed(self, client: sa_asyncio.AsyncSession) -> bool:
        # sqlalchemy session is never truly closed
        return await self._internal_adapter.is_closed(client.bind)

    async def begin(self, client: sa_asyncio.AsyncSession) -> None:
        await self._internal_adapter.begin(client.bind)

    async def commit(self, client: sa_asyncio.AsyncSession) -> None:
        await self._internal_adapter.commit(client.bind)

    async def rollback(self, client: sa_asyncio.AsyncSession) -> None:
        await self._internal_adapter.rollback(client.bind)

    async def in_atomic(self, client: sa_asyncio.AsyncSession) -> bool:
        return client.in_transaction()

    def context(
        self,
    ) -> "AsyncSessionContext":
        return AsyncSessionContext(self)



import os
from os import environ
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Iterator
from typing import MutableMapping
from typing import TypeVar
from typing import Union
from typing import overload

from gyver.exc import InvalidCast
from gyver.exc import MissingName
from gyver.utils.exc import panic

T = TypeVar("T")


class EnvMapping(MutableMapping[str, str]):
    def __init__(self, mapping: MutableMapping[str, str] = environ) -> None:
        self._mapping = mapping
        self._already_read = set[str]()

    def __getitem__(self, name: str):
        val = self._mapping[name]
        self._already_read.add(name)
        return val

    def __setitem__(self, name: str, value: str):
        if name in self._already_read:
            raise panic(KeyError, f"{name} already read, cannot change its value")
        self._mapping[name] = value

    def __delitem__(self, name: str) -> None:
        if name in self._already_read:
            raise panic(KeyError, f"{name} already read, cannot delete")
        del self._mapping[name]

    def __iter__(self) -> Iterator[str]:
        yield from self._mapping

    def __len__(self) -> int:
        return len(self._mapping)


default_mapping = EnvMapping()


class MISSING:
    pass


def _default_cast(a: Any):
    return a


class Config:
    def __init__(
        self,
        env_file: Union[str, Path, None] = None,
        mapping: EnvMapping = default_mapping,
    ) -> None:
        self._mapping = mapping
        self._file_values = {}
        if env_file and os.path.isfile(env_file):
            self._read_file(env_file)

    def _read_file(self, env_file: Union[str, Path]):
        with open(env_file) as buf:
            for line in buf:
                if line.startswith("#"):
                    continue
                name, value = line.split("=", 1)
                self._file_values[name.strip()] = value.strip()

    def _cast(self, name: str, val: Any, cast: Callable) -> Any:
        try:
            val = cast(val)
        except Exception as e:
            raise panic(InvalidCast, f"{name} received an invalid value {val}") from e
        else:
            return val

    def _get_val(
        self, name: str, default: Union[Any, type[MISSING]] = MISSING
    ) -> Union[Any, type[MISSING]]:
        return self._mapping.get(name, self._file_values.get(name, default))

    def get(
        self,
        name: str,
        cast: Callable = _default_cast,
        default: Union[Any, type[MISSING]] = MISSING,
    ) -> Any:
        val = self._get_val(name, default)
        if val is MISSING:
            raise panic(MissingName, f"{name} not found and no default was given")
        return self._cast(name, val, cast)

    @overload
    def __call__(
        self,
        name: str,
        cast: Union[Callable[[Any], T], type[T]] = _default_cast,
        default: type[MISSING] = MISSING,
    ) -> T:
        ...

    @overload
    def __call__(
        self,
        name: str,
        cast: Union[Callable[[Any], T], type[T]] = _default_cast,
        default: T = ...,
    ) -> T:
        ...

    def __call__(
        self,
        name: str,
        cast: Union[Callable[[Any], T], type[T]] = _default_cast,
        default: Union[T, type[MISSING]] = MISSING,
    ) -> T:
        return self.get(name, cast, default)



from . import comp
from ._helpers import retrieve_attr
from .exc import FieldNotFound
from .exc import FilterError
from .interface import ApplyClause
from .interface import BindClause
from .interface import Mapper
from .order_by import OrderBy
from .order_by import OrderDirection
from .paginate import FieldPaginate
from .paginate import LimitOffsetPaginate
from .paginate import Paginate
from .utils import as_date
from .utils import as_lower
from .utils import as_time
from .utils import as_upper
from .where import FieldResolver
from .where import Where
from .where import ApplyWhere
from .where import and_
from .where import or_

__all__ = [
    "Where",
    "ApplyWhere",
    "FieldResolver",
    "and_",
    "or_",
    "OrderBy",
    "OrderDirection",
    "as_date",
    "as_time",
    "as_lower",
    "as_upper",
    "FieldPaginate",
    "LimitOffsetPaginate",
    "Paginate",
    "FilterError",
    "FieldNotFound",
    "comp",
    "BindClause",
    "ApplyClause",
    "retrieve_attr",
    "Mapper",
]



from urllib.parse import quote

from gyver.url.encode import Encodable
from gyver.url.utils import is_valid_encoded_path
from gyver.url.utils import utf8


class Fragment(Encodable):
    def __init__(self, fragment_str: str) -> None:
        self.set(fragment_str)

    def encode(self):
        return self.fragment_str

    def set(self, fragment_str: str):
        if not is_valid_encoded_path.match(fragment_str):
            fragment_str = quote(utf8(fragment_str))
        self.fragment_str = fragment_str
        return self

    def __str__(self):
        return self.fragment_str



from typing import Any
from typing import Generator
from typing import Sequence
from typing import Union

from attr._make import Factory
from attrs import NOTHING
from attrs import Attribute

from gyver.attrs import define
from gyver.config.adapter.interface import FieldResolverStrategy
from gyver.config.config import MISSING
from gyver.exc import InvalidCast
from gyver.utils import panic


@define
class AttrsResolverStrategy(FieldResolverStrategy[Attribute]):
    field: Attribute

    def cast(self) -> type:
        if field_type := self.field.type:
            return field_type
        raise panic(InvalidCast, f"Field {self.field.name} has no valid type")

    def names(self) -> Sequence[str]:
        return tuple(
            item for item in (self.field.name, self.field.alias) if item is not None
        )

    def init_name(self) -> str:
        return self.field.alias or self.field.name

    def default(self) -> Union[Any, type[MISSING]]:
        default = self.field.default
        if default is None or default in (Ellipsis, NOTHING):
            return MISSING
        return default.factory() if isinstance(default, Factory) else default

    @staticmethod
    def iterfield(config_class: type) -> Generator[Attribute, Any, Any]:
        yield from config_class.__attrs_attrs__  # type: ignore



from gyver.url import URL
from gyver.url import Netloc

from .config import CacheConfig


def make_uri(config: CacheConfig) -> str:
    """
    The make_uri function creates a Redis URI from the given CacheConfig.

    :param config: CacheConfig: Pass the configuration to the function
    :return: A string that is a valid uri for connecting to redis
    """
    url = URL("")
    url.scheme = "redis"
    url.add(
        netloc_args=Netloc("").set(
            config.host, config.user, config.password, config.port
        )
    )
    return url.encode()



from .asyncio import AsyncConnectionAdapter
from .asyncio import AsyncSaContext
from .asyncio import AsyncSessionContext
from .sync import SaAdapter
from .sync import SaContext
from .sync import SessionContext

__all__ = [
    "AsyncSaContext",
    "AsyncSessionContext",
    "AsyncConnectionAdapter",
    "SaAdapter",
    "SaContext",
    "SessionContext",
]



from enum import Enum
from typing import NamedTuple

from typing_extensions import Self


class EnvTuple(NamedTuple):
    val: str
    weight: int


class Env(EnvTuple, Enum):
    LOCAL = "local", 1
    TEST = "test", 2
    DEV = "dev", 3
    PRD = "prd", 4

    @property
    def value(self) -> EnvTuple:
        return super().value

    @property
    def val(self):
        return self.value.val

    @property
    def weight(self):
        return self.value.weight

    @classmethod
    def new(cls, val: str) -> Self:
        return next(value for value in cls if value.val == val)



import os
from dataclasses import dataclass
from pathlib import Path
from typing import Union

from gyver import utils

from .config import Config
from .config import EnvMapping
from .config import default_mapping
from .typedef import Env


@dataclass(frozen=True)
class DotFile:
    filename: Union[str, Path]
    env: Env
    apply_to_lower: bool = False

    def __gt__(self, other: "DotFile") -> bool:
        return self.env.weight > other.env.weight

    def is_higher(self, env: Env) -> bool:
        return self.env.weight >= env.weight


class EnvConfig(Config):
    def __init__(
        self,
        *dotfiles: DotFile,
        env_var: str = "CONFIG_ENV",
        mapping: EnvMapping = default_mapping
    ) -> None:
        self._env_var = env_var
        self._dotfiles = dotfiles
        self._mapping = mapping
        self._file_values = {}
        if self.dotfile:
            self._read_file(self.dotfile.filename)

    @utils.lazyfield
    def env(self):
        return self.__call__(self._env_var, Env.new)

    @utils.lazyfield
    def dotfile(self):
        """
        The dotfile function returns the dotfile object that is applicable to
        the current environment.  It traverses the list of dotfiles in reverse order,
        so that higher priority dotfiles are checked first.  If no matching dotfile is
        found, it returns None.

        :return: The first dotfile that is not higher than the current environment
        """

        for dot in sorted(self._dotfiles, reverse=True):
            if not dot.is_higher(self.env):
                break
            if dot.env != self.env and (
                not dot.apply_to_lower or not dot.is_higher(self.env)
            ):
                continue
            if not os.path.isfile(dot.filename):
                continue
            return dot



import contextlib
import typing

import sqlalchemy as sa

from gyver.attrs import define
from gyver.database.typedef import ClauseType

from . import _helpers
from . import comp as cp
from . import interface

T = typing.TypeVar("T")

CACHE_MAXLEN = 250


class _BindCache:
    def __init__(self) -> None:
        self._cached: dict[typing.Hashable, interface.Comparison] = {}
        self.maxlen = CACHE_MAXLEN

    def get(self, key: typing.Hashable) -> typing.Optional[interface.Comparison]:
        try:
            return self._cached.get(key)
        except TypeError:
            return None

    def set(
        self, key: typing.Hashable, value: interface.Comparison
    ) -> interface.Comparison:
        if len(self._cached) >= CACHE_MAXLEN:
            self._cached.pop(tuple(self._cached)[0])
        with contextlib.suppress(TypeError):
            self._cached[key] = value
        return value

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.clear()

    def clear(self):
        self._cached.clear()


_cache = _BindCache()


@define
class Resolver(typing.Generic[T]):
    val: typing.Any

    def resolve(self, mapper: interface.Mapper):
        del mapper
        val = self.val
        if isinstance(val, list):
            return tuple(val)
        elif isinstance(val, set):
            return frozenset(val)
        return val

    def __bool__(self):
        return self.val is not None


class FieldResolver(Resolver[str]):
    def resolve(self, mapper: interface.Mapper):
        return _helpers.retrieve_attr(mapper, self.val)


@define(frozen=False)
class Where(interface.BindClause, typing.Generic[T]):
    field: str
    expected: Resolver[T]
    comp: interface.Comparator[T] = cp.equals

    def __init__(
        self,
        field: str,
        expected: typing.Optional[T] = None,
        comp: interface.Comparator[T] = cp.equals,
        resolver_class: type[Resolver[T]] = Resolver,
    ) -> None:
        self.__gattrs_init__(  # type: ignore
            field,
            resolver_class(expected),
            comp,
        )

    type_ = ClauseType.BIND

    def bind(self, mapper: interface.Mapper) -> interface.Comparison:
        if self.comp is cp.always_true:
            return AlwaysTrue().bind(mapper)
        resolved = self.expected.resolve(mapper)
        if not isinstance(mapper, typing.Hashable):
            return self._get_comparison(
                _helpers.retrieve_attr(mapper, self.field),
                resolved,
            )
        if (value := _cache.get((mapper, self.field, resolved, self.comp))) is not None:
            return value
        attr = _helpers.retrieve_attr(mapper, self.field)

        return _cache.set(
            (mapper, self.field, resolved, self.comp),
            self._get_comparison(attr, resolved),
        )

    def _get_comparison(self, attr: interface.FieldType, resolved: typing.Any):
        return self.comp(attr, resolved) if self.expected else sa.true()


class _JoinBind(interface.BindClause):
    def __init__(
        self,
        items: typing.Sequence[interface.BindClause],
        operator: typing.Callable[..., interface.Comparison],
    ) -> None:
        self.items = items
        self.operator = operator

    type_ = ClauseType.BIND

    def __bool__(self):
        return bool(self.items)

    def bind(self, mapper: interface.Mapper) -> interface.Comparison:
        return (
            self.operator(*(item.bind(mapper) for item in self.items))
            if self
            else sa.true()
        )


def and_(*bind: interface.BindClause) -> _JoinBind:
    return _JoinBind(bind, sa.and_)


def or_(*bind: interface.BindClause) -> _JoinBind:
    return _JoinBind(bind, sa.or_)


_placeholder_column = sa.Column("placeholder")


class AlwaysTrue(interface.BindClause):
    type_ = ClauseType.BIND

    def bind(self, mapper: interface.Mapper) -> interface.Comparison:
        return cp.always_true(_placeholder_column, mapper)


class RawQuery(interface.BindClause):
    type_ = ClauseType.BIND

    def __init__(self, cmp: interface.Comparison) -> None:
        self._cmp = cmp

    def bind(self, mapper: interface.Mapper) -> interface.Comparison:
        del mapper
        return self._cmp


class ApplyWhere(interface.ApplyClause):
    type_ = ClauseType.APPLY

    def __init__(self, mapper: interface.Mapper, *where: interface.BindClause) -> None:
        self.where = and_(*where).bind(mapper)

    def apply(self, query: interface.ExecutableT) -> interface.ExecutableT:
        return query.where(self.where)



import asyncio
import contextlib
import threading
import typing

from gyver.utils import lazyfield

from . import interfaces
from .typedef import T


class Context(typing.Generic[T]):
    def __init__(self, adapter: interfaces.Adapter[T]) -> None:
        """
        Initialize a new Context.
        :param adapter: An adapter that will be used to acquire and release resources.
        """
        self._adapter = adapter
        self._stack = 0  # Keeps track of how many frames are using this context
        self._lock = threading.Lock()  # A lock to ensure thread safety

    @lazyfield
    def client(self) -> T:
        """
        Returns the current resource being used by the context.
        Acquires a new resource if the current one is closed or doesn't exist.
        """
        return self.adapter.new()

    @property
    def stack(self):
        """
        Returns how many frames are using this context
        """
        return self._stack

    @property
    def adapter(self) -> interfaces.Adapter[T]:
        """
        Returns the adapter that is being used by this context
        """
        return self._adapter

    def is_active(self) -> bool:
        """
        Returns whether the context is currently in use
        """
        return self._stack > 0

    def acquire(self):
        """
        Acquires a new resource from the adapter and increases the stack count.
        """
        with self._lock:
            if self.adapter.is_closed(self.client):
                Context.client.cleanup(self)
            self._stack += 1
            return self.client

    def release(self):
        """
        Releases the current resource if the stack count is 1,
        and decreases the stack count.
        """
        with self._lock:
            if self._stack == 1:
                self.adapter.release(self.client)
                Context.client.cleanup(self)
            self._stack -= 1

    @contextlib.contextmanager
    def open(self):
        """
        A context manager that acquires and releases resources without returning it.
        """
        with self:
            yield

    @contextlib.contextmanager
    def begin(self):
        """
        A context manager that acquires and releases resources and returns it.
        """
        with self as client:
            yield client

    def __enter__(self):
        """
        Acquires a new resource from the adapter and increases the stack count.
        """
        return self.acquire()

    def __exit__(self, *_):
        """
        Releases the current resource if the stack count is 1,
         and decreases the stack count.
        """
        self.release()


class AsyncContext(typing.Generic[T]):
    def __init__(self, adapter: interfaces.AsyncAdapter[T]) -> None:
        """
        Initialize a new AsyncContext.
        :param adapter: An async adapter that will be used to acquire
        and release resources.
        """
        self._adapter = adapter
        self._stack = 0  # Keeps track of how many frames are using this context
        self._client: typing.Optional[T] = None  # The current resource being used
        self._lock = asyncio.Lock()  # A lock to ensure thread safety

    @property
    def stack(self) -> int:
        """
        Returns how many frames are using this context
        """
        return self._stack

    @property
    def adapter(self) -> interfaces.AsyncAdapter[T]:
        """
        Returns the adapter that is being used by this context
        """
        return self._adapter

    def is_active(self) -> bool:
        """
        Returns whether the context is currently in use
        """
        return self._stack > 0

    async def client(self) -> T:
        """
        Returns the current resource being used by the context.
        Acquires a new resource if the current one is closed or doesn't exist.
        """
        if self._client is None or await self._adapter.is_closed(self._client):
            self._client = await self._adapter.new()
        return self._client

    async def acquire(self):
        """
        Acquires a new resource from the adapter and increases the stack count.
        """
        async with self._lock:
            client = await self.client()
            self._stack += 1
            return client

    async def release(self):
        """
        Releases the current resource if the stack count is 1,
        and decreases the stack count.
        """
        async with self._lock:
            if self._client is None:
                raise RuntimeError(
                    "Release should not be called before client initialization"
                )
            if self._stack == 1:
                await self.adapter.release(self._client)
                self._client = None
            self._stack -= 1

    @contextlib.asynccontextmanager
    async def open(self):
        """
        An async context manager that acquires
        and releases resources without returning it.
        """
        async with self:
            yield

    @contextlib.asynccontextmanager
    async def begin(self):
        """
        An async context manager that acquires and releases resources and returns it.
        """
        async with self as client:
            yield client

    async def __aenter__(self):
        """
        Acquires a new resource from the adapter and increases the stack count.
        """
        return await self.acquire()

    async def __aexit__(self, *_):
        """
        Releases the current resource if the stack count
        is 1, and decreases the stack count.
        """
        await self.release()



import re
from typing import Callable
from typing import TypeVar
from typing import Union

from typing_extensions import Concatenate
from typing_extensions import ParamSpec

T = TypeVar("T")
R = TypeVar("R")
P = ParamSpec("P")


def try_cast(
    val: T, cast: Callable[Concatenate[T, P], R], *args: P.args, **kwargs: P.kwargs
) -> Union[T, R]:
    try:
        return cast(val, *args, **kwargs)
    except Exception:
        return val


def utf8(val: str):
    return try_cast(val, str.encode, "utf-8")


PERCENT_REGEX = r"\%[a-fA-F\d][a-fA-F\d]"


is_valid_encoded_path = re.compile(
    r"^([\w%s]|(%s))*$" % (re.escape("-.~:@!$&'()*+,;="), PERCENT_REGEX)
)



import functools
import warnings
from typing import Callable
from typing import TypeVar
from typing import cast

from typing_extensions import ParamSpec

from gyver.attrs import define

from .exc import panic


def _not_implemented(msg: str):
    def inner(*_, **__kwds__):
        raise panic(NotImplementedError, msg)

    return inner


P = ParamSpec("P")
T = TypeVar("T")


def cache(f: Callable[P, T]) -> Callable[P, T]:
    return cast(Callable[P, T], functools.cache(f))


def deprecated(func: Callable[P, T]) -> Callable[P, T]:
    @functools.wraps(func)
    def inner(*args: P.args, **kwargs: P.kwargs) -> T:
        if not hasattr(func, "__warn_deprecated__"):
            warnings.warn(
                f"{func.__qualname__} is "
                "deprecated and can be removed without notice",
                DeprecationWarning,
            )
            func.__warn_deprecated__ = True
        return func(*args, **kwargs)

    return inner


frozen = deprecated(define)


class DeprecatedClass:
    __warn_deprecated__ = False

    def __init__(self) -> None:
        if not type(self).__warn_deprecated__:
            warnings.warn(
                f"{'.'.join((type(self).__module__, type(self).__name__))} is "
                "deprecated and can be removed without notice",
                DeprecationWarning,
            )
            type(self).__warn_deprecated__ = True
